/*
	USB mass storage driver for BeOS R5
	Copyright (C) 2000 ITO, Takayuki
	All rights reserved
*/

#include "driver.h"

#define	BULKONLY_RESET	0xff
#define	GET_MAX_LUN	0xfe

#define	DPRINTF_INFO(x)	dprintf x
#define	DPRINTF_ERR(x)	dprintf x

struct bulkonly_cbw
{
	char	dCBWSignature [4];		/* "USBC" */
	uint32	dCBWTag;
	uint32	dCBWDataXferLength;
	/* offset 12 */
	uint8	reserved_12:7;
	uint8	Direction:1;

	uint8	bCBWLUN;
	uint8	bCBWCBLength;
	uint8	CBWCB [16];
} __attribute__ ((packed));

struct bulkonly_csw
{
	char	dCSWSignature [4];		/* "USBS" */
	uint32	dCSWTag;
	uint32	dCSWDataResidue;
	uint8	bCSWStatus;				/* USB_MASS_STATUS_* */
} __attribute__ ((packed));

static status_t bulkonly_exec_cmd_dataio_sg (
	my_device_info *my_dev, 
	const void *cdb, size_t cdblen, 
	const iovec *vec, size_t vecsiz, 
	size_t length, 
	bool isOutput)
{
	status_t st;
	struct bulkonly_cbw cbw;
	struct bulkonly_csw csw;
	int i;

	assert (my_dev != NULL && cdb != NULL && cdblen > 0);
	my_dev->cmd_status = 0;

	/* send CBW */

	memset (&cbw, 0, sizeof cbw);
	strncpy (cbw.dCBWSignature, "USBC", 4);
	cbw.dCBWDataXferLength = length;
	cbw.Direction = isOutput ? 0 : 1;
	cbw.bCBWCBLength = cdblen;
	memcpy (cbw.CBWCB, cdb, cdblen);
	st = do_bulk (my_dev, my_dev->bulkOutEpt, &cbw, sizeof cbw);
#ifdef DATA_TOGGLE_WORKAROUND
	if (my_dev->actual_length > 0)	/* FIXME */
		my_dev->bulkOutEptToggle ^= 1;
#endif
	if (st < 0)
	{
		DPRINTF_ERR ((MY_ID "bulkonly_exec_cmd_dataio_sg(): command out failed %d %d\n",
			(int) my_dev->usbd_status, (int) my_dev->bus_status));
		if (my_dev->bus_status == B_USB_STATUS_DEVICE_STALLED)
			usb->clear_feature (my_dev->bulkInEpt->handle, 
				USB_FEATURE_ENDPOINT_HALT);
		return B_ERROR;
	}
	DPRINTF_INFO ((MY_ID "bulkonly_exec_cmd_dataio_sg(): command sent\n"));

	/* send/recv data */

	if (vec != NULL && vecsiz > 0 && length > 0)
	{
		const usb_endpoint_info *ept;
#ifdef DATA_TOGGLE_WORKAROUND
		int maxPacketSize, toggle;
#endif

		for (i = 0; i < vecsiz; i++)
			assert (vec [i].iov_base != NULL);
		if (isOutput)
			ept = my_dev->bulkOutEpt;
		else
			ept = my_dev->bulkInEpt;
		st = do_bulk_sg (my_dev, ept, vec, vecsiz, length);
		my_dev->bulk_length = my_dev->actual_length;

#ifdef DATA_TOGGLE_WORKAROUND
		maxPacketSize = ept->descr->max_packet_size;
		if (isOutput)
			toggle = my_dev->bulkOutEptToggle;
		else
			toggle = my_dev->bulkInEptToggle;
		toggle = (toggle + (my_dev->actual_length + maxPacketSize - 1) / 
			maxPacketSize) & 1;
		if (isOutput)
			my_dev->bulkOutEptToggle = toggle;
		else
			my_dev->bulkInEptToggle = toggle;
#endif

		if (st < 0)
		{
			DPRINTF_ERR ((MY_ID "bulkonly_exec_cmd_dataio_sg(): data i/o failed %d %d\n",
				(int) my_dev->usbd_status, (int) my_dev->bus_status));
			if (my_dev->bus_status == B_USB_STATUS_DEVICE_STALLED)
				usb->clear_feature (ept->handle, USB_FEATURE_ENDPOINT_HALT);
			// endpoint stall is not fatal
		}
		else
			DPRINTF_INFO ((MY_ID "bulkonly_exec_cmd_dataio_sg(): data i/o completed\n"));
	}

	/* recv CSW */

	st = do_bulk (my_dev, my_dev->bulkInEpt, &csw, sizeof csw);
#ifdef DATA_TOGGLE_WORKAROUND
	if (my_dev->actual_length > 0)	/* FIXME */
		my_dev->bulkInEptToggle ^= 1;
#endif
	if (st < 0)
	{
		DPRINTF_ERR ((MY_ID "bulkonly_exec_cmd_dataio_sg(): status in failed %d %d\n",
			(int) my_dev->usbd_status, (int) my_dev->bus_status));
		if (my_dev->bus_status == B_USB_STATUS_DEVICE_STALLED)
			usb->clear_feature (my_dev->bulkInEpt->handle, 
				USB_FEATURE_ENDPOINT_HALT);
		// endpoint stall is not fatal
		return B_ERROR;
	}
	DPRINTF_INFO ((MY_ID "bulkonly_exec_cmd_dataio_sg(): status received\n"));
	if (strncmp (csw.dCSWSignature, "USBS", 4) != 0 ||
		csw.dCSWTag != cbw.dCBWTag)
		DPRINTF_ERR ((MY_ID "warning: invalid CSW\n"));
	my_dev->cmd_status = csw.bCSWStatus;
	if (my_dev->bulk_length != cbw.dCBWDataXferLength - csw.dCSWDataResidue)
	{
		DPRINTF_ERR ((MY_ID "warning: %u != %u - %u\n",
			(uint) my_dev->bulk_length, (uint) cbw.dCBWDataXferLength,
			(uint)csw.dCSWDataResidue));
		my_dev->cmd_status = USB_MASS_STATUS_PHASE_ERR;
	}

#ifdef DATA_TOGGLE_WORKAROUND
	DPRINTF_INFO ((MY_ID "toggles=%d %d\n", my_dev->bulkInEptToggle,
		my_dev->bulkOutEptToggle));
#endif

	return (my_dev->cmd_status == USB_MASS_STATUS_PASSED) ? B_OK : B_ERROR;
}

static status_t bulkonly_init (my_device_info *my_dev)
{
	return B_ERROR;
}

static status_t bulkonly_finish (my_device_info *my_dev)
{
	return B_ERROR;
}

static status_t bulkonly_exec_cmd_nondata (
	my_device_info *my_dev, 
	const void *cdb, size_t cdblen)
{
	DPRINTF_INFO ((MY_ID "bulkonly_exec_cmd_nondata()\n"));
	return bulkonly_exec_cmd_dataio_sg 
		(my_dev, cdb, cdblen, NULL, 0, 0, false);
}

static status_t bulkonly_exec_cmd_datain_sg (
	my_device_info *my_dev, 
	const void *cdb, size_t cdblen, 
	const iovec *vec, size_t vecsiz, 
	size_t length)
{
	DPRINTF_INFO ((MY_ID "bulkonly_exec_cmd_datain_sg()\n"));
	return bulkonly_exec_cmd_dataio_sg 
		(my_dev, cdb, cdblen, vec, vecsiz, length, false);
}

static status_t bulkonly_exec_cmd_dataout_sg (
	my_device_info *my_dev, 
	const void *cdb, size_t cdblen,
	const iovec *vec, size_t vecsiz, 
	size_t length)
{
	DPRINTF_INFO ((MY_ID "bulkonly_exec_cmd_dataout_sg()\n"));
	return bulkonly_exec_cmd_dataio_sg 
		(my_dev, cdb, cdblen, vec, vecsiz, length, true);
}

static status_t bulkonly_exec_cmd_datain (
	my_device_info *my_dev, 
	const void *cdb, size_t cdblen, 
	void *buf, size_t buflen)
{
	iovec vec;
	DPRINTF_INFO ((MY_ID "bulkonly_exec_cmd_datain()\n"));
	vec.iov_base = buf;
	vec.iov_len = buflen;
	return bulkonly_exec_cmd_dataio_sg 
		(my_dev, cdb, cdblen, &vec, 1, buflen, false);
}

static status_t
bulkonly_exec_cmd_dataout (
	my_device_info *my_dev, 
	const void *cdb, size_t cdblen,
	const void *buf, size_t buflen)
{
	iovec vec;
	DPRINTF_INFO ((MY_ID "bulkonly_exec_cmd_dataout()\n"));
	vec.iov_base = (void *) buf;
	vec.iov_len = buflen;
	return bulkonly_exec_cmd_dataio_sg 
		(my_dev, cdb, cdblen, &vec, 1, buflen, true);
}

transport_hooks bulkonly_hooks =
{
	bulkonly_init,
	bulkonly_exec_cmd_nondata,
	bulkonly_exec_cmd_datain,
	bulkonly_exec_cmd_dataout,
	bulkonly_exec_cmd_datain_sg,
	bulkonly_exec_cmd_dataout_sg,
	bulkonly_finish
};
