/*
	USB mass storage driver for BeOS R5
	Copyright (C) 2000 ITO, Takayuki
	All rights reserved
*/

#include "driver.h"

#define	DPRINTF_ERR(x)	dprintf x
#define	DPRINTF_INFO(x)	((void)0)

#define	ADSC	0x00

static status_t cbi_exec_cmd_dataio_sg (
	my_device_info *my_dev, 
	const void *cdb, size_t cdblen, 
	const iovec *vec, size_t vecsiz, 
	size_t length, 
	bool isOutput)
{
	status_t st;
	uint8 cmd_stat [2];

	assert (my_dev != NULL && cdb != NULL && cdblen > 0);
	my_dev->cmd_status = USB_MASS_STATUS_PASSED;

	/* send command */

	st = do_control (my_dev, USB_REQTYPE_CLASS | USB_REQTYPE_INTERFACE_OUT, 
		ADSC, 0, DEFAULT_INTERFACE, cdblen, (void *) cdb);
	if (st < 0)
	{
		DPRINTF_ERR ((MY_ID MY_ERR "cbi_exec_cmd_dataio_sg(): command out failed %d %d\n",
			(int) my_dev->usbd_status, (int) my_dev->bus_status));
		/* device may stall control pipe to indicate error */
		if (my_dev->bus_status == B_USB_STATUS_DEVICE_STALLED)
			usb->clear_feature (my_dev->bulkInEpt->handle, 
				USB_FEATURE_ENDPOINT_HALT);
		return st;
	}

	/* send/recv data */

	if (vec != NULL && vecsiz > 0 && length > 0)
	{
		const usb_endpoint_info *ept;
		int i;
#ifdef DATA_TOGGLE_WORKAROUND
		int maxPacketSize, toggle;
#endif

		for (i = 0; i < vecsiz; i++)
			assert (vec [i].iov_base != NULL);
		if (isOutput)
			ept = my_dev->bulkOutEpt;
		else
			ept = my_dev->bulkInEpt;
		assert (ept != NULL);
		st = do_bulk_sg (my_dev, ept, vec, vecsiz, length);
		my_dev->bulk_length = my_dev->actual_length;

#ifdef DATA_TOGGLE_WORKAROUND
		maxPacketSize = ept->descr->max_packet_size;
		if (isOutput)
			toggle = my_dev->bulkOutEptToggle;
		else
			toggle = my_dev->bulkInEptToggle;
		toggle = (toggle + (my_dev->actual_length + maxPacketSize - 1) / 
			maxPacketSize) & 1;
		if (isOutput)
			my_dev->bulkOutEptToggle = toggle;
		else
			my_dev->bulkInEptToggle = toggle;
#endif

		if (st < 0)
		{
			DPRINTF_ERR ((MY_ID MY_ERR "cbi_exec_cmd_dataio_sg(): data i/o failed %d %d\n",
				(int) my_dev->usbd_status, (int) my_dev->bus_status));
			/* device may stall bulk pipe to indicate error */
			if (my_dev->bus_status == B_USB_STATUS_DEVICE_STALLED)
				usb->clear_feature (ept->handle, USB_FEATURE_ENDPOINT_HALT);
			return st;
		}
	}

	/* recv status */

	if (my_dev->protocol == USB_PROTOCOL_MASS_CB)
		/* command error, if any, may be already shown 
		by control pipe stall or bulk pipe stall */
		return B_OK;

	st = do_interrupt (my_dev, my_dev->intrEpt, cmd_stat, 2);
#ifdef DATA_TOGGLE_WORKAROUND
	if (my_dev->actual_length > 0)
	{
		my_dev->intrEptToggle ^= 1;
		DPRINTF_INFO ((MY_ID "toggles=%d %d %d\n", my_dev->bulkInEptToggle,
			my_dev->bulkOutEptToggle, my_dev->intrEptToggle));
	}
#endif

	if (st < 0)
	{
		DPRINTF_ERR ((MY_ID MY_ERR "cbi_exec_cmd_dataio(): status in failed %d %d\n",
			(int) my_dev->usbd_status, (int) my_dev->bus_status));
		if (my_dev->bus_status == B_USB_STATUS_DEVICE_STALLED)
			usb->clear_feature (my_dev->intrEpt->handle, 
				USB_FEATURE_ENDPOINT_HALT);
		return st;
	}

	if (my_dev->subclass == USB_SUBCLASS_MASS_UFI)
		/* ASC & ASCQ */
		my_dev->cmd_status = (cmd_stat [0] == 0 && cmd_stat [1] == 0) ? 
			USB_MASS_STATUS_PASSED : USB_MASS_STATUS_FAILED;
	else
		/* 0:Pass, 1:Fail, 2:PhaseError */
		my_dev->cmd_status = cmd_stat [1] & 3;

	return (my_dev->cmd_status == USB_MASS_STATUS_PASSED) ? B_OK : B_ERROR;
}

static status_t cbi_init (my_device_info *my_dev)
{
	assert (my_dev != NULL);
	return B_ERROR;
}

static status_t cbi_finish (my_device_info *my_dev)
{
	assert (my_dev != NULL);
	return B_ERROR;
}

static status_t cbi_exec_cmd_nondata (
	my_device_info *my_dev, 
	const void *cdb, size_t cdblen)
{
	return cbi_exec_cmd_dataio_sg (my_dev, cdb, cdblen, NULL, 0, 0, false);
}

static status_t cbi_exec_cmd_datain_sg (
	my_device_info *my_dev, 
	const void *cdb, size_t cdblen,
	const iovec *vec, size_t vecsiz, 
	size_t length)
{
	return cbi_exec_cmd_dataio_sg (my_dev, cdb, cdblen, vec, vecsiz, length, false);
}

static status_t cbi_exec_cmd_dataout_sg (
	my_device_info *my_dev, 
	const void *cdb, size_t cdblen,
	const iovec *vec, size_t vecsiz, 
	size_t length)
{
	return cbi_exec_cmd_dataio_sg (my_dev, cdb, cdblen, vec, vecsiz, length, true);
}

static status_t cbi_exec_cmd_datain (
	my_device_info *my_dev, 
	const void *cdb, size_t cdblen,
	void *buf, size_t buflen)
{
	iovec vec;
	vec.iov_base = buf;
	vec.iov_len = buflen;
	return cbi_exec_cmd_dataio_sg (my_dev, cdb, cdblen, &vec, 1, buflen, false);
}

static status_t cbi_exec_cmd_dataout (
	my_device_info *my_dev, 
	const void *cdb, size_t cdblen,
	const void *buf, size_t buflen)
{
	iovec vec;
	vec.iov_base = (void *) buf;
	vec.iov_len = buflen;
	return cbi_exec_cmd_dataio_sg (my_dev, cdb, cdblen, &vec, 1, buflen, true);
}

transport_hooks cbi_hooks =
{
	cbi_init,
	cbi_exec_cmd_nondata,
	cbi_exec_cmd_datain,
	cbi_exec_cmd_dataout,
	cbi_exec_cmd_datain_sg,
	cbi_exec_cmd_dataout_sg,
	cbi_finish
};
