/*
	USB mass storage driver for BeOS R5
	Copyright (C) 2000 ITO, Takayuki
	All rights reserved
*/

#include "driver.h"

#define	DPRINTF_INFO(x)	dprintf x
#define	DPRINTF_ERR(x)	dprintf x

#define	CLEAR_FEATURE	1
#define	ENDPOINT_HALT	0

#define	MAX_CDB_LEN	16	/* SCSI-3 */
#define	INQUIRY_LEN	36
#define	SENSE_LEN	18
#define	MODE_HDR_LEN	8

static void build_cdb6 (
	uint8 *cdb, 
	int op, 
	int lun, 
	uint32 lba, 
	uint32 len, 
	int ctrl)
{
	assert (cdb != NULL);
	if ((op >> 5) != 0)
		DPRINTF_ERR ((MY_ID MY_WARN "CDB op 0x%x is not 6-byte command\n", op));
	cdb [0] = op;
	cdb [1] = (lun << 5) | (lba >> 16);
	cdb [2] = lba >>  8;
	cdb [3] = lba      ;
	cdb [4] = len;
	cdb [5] = ctrl;
}

static void build_cdb10 (
	uint8 *cdb, 
	int op, 
	int lun, 
	uint32 lba, 
	uint32 len, 
	int ctrl)
{
	assert (cdb != NULL);
	if ((op >> 5) != 1 && (op >> 5) != 2)
		DPRINTF_ERR ((MY_ID MY_WARN "CDB op 0x%x is not 10-byte command\n", op));
	cdb [0] = op;
	cdb [1] = lun << 5;
	cdb [2] = lba >> 24;
	cdb [3] = lba >> 16;
	cdb [4] = lba >>  8;
	cdb [5] = lba      ;
	cdb [6] = 0;	/* reserved */
	cdb [7] = len >>  8;
	cdb [8] = len      ;
	cdb [9] = ctrl;
}

static void build_cdb12 (
	uint8 *cdb, 
	int op, 
	int lun, 
	uint32 lba, 
	uint32 len, 
	int ctrl)
{
	assert (cdb != NULL);
	if ((op >> 5) != 5)
		DPRINTF_ERR ((MY_ID MY_WARN "CDB op 0x%x is not 12-byte command\n", op));
	cdb [0] = op;
	cdb [1] = lun << 5;
	cdb [2] = lba >> 24;
	cdb [3] = lba >> 16;
	cdb [4] = lba >>  8;
	cdb [5] = lba      ;
	cdb [6] = len >> 24;
	cdb [7] = len >> 16;
	cdb [8] = len >>  8;
	cdb [9] = len      ;
	cdb [10] = 0;	/* reserved */
	cdb [11] = ctrl;
}

static int build_cdb (
	uint8 *cdb, 
	int op, 
	int lun, 
	uint32 lba, 
	uint32 len, 
	int ctrl)
{
	int cdblen = 0;

	memset (cdb, 0, MAX_CDB_LEN);
	switch (op >> 5)
	{
	case 0:
		build_cdb6 (cdb, op, lun, lba, len, ctrl);
		cdblen = 6;
		break;
	case 1:
	case 2:
		build_cdb10 (cdb, op, lun, lba, len, ctrl);
		cdblen = 10;
		break;
	case 3:
		/* reserved */
		break;
	case 4:
		/* SCSI-3 */
//		build_cdb16 (cdb, op, lun, lba, len, ctrl);
		cdblen = 16;
		break;
	case 5:
		build_cdb12 (cdb, op, lun, lba, len, ctrl);
		cdblen = 12;
		break;
	case 6:
	case 7:
		/* vendor specific */
		break;
	}

	return cdblen;
}

static status_t conv_error (
	my_device_info *my_dev, 
	bool autosense)
{
	status_t ret;
	sense_data	sense;

	/* see USB bus manager error */

	switch (my_dev->usbd_status)
	{
	case B_USBD_SUCCESS:
		ret = B_OK;
		break;

	case B_USBD_BAD_HANDLE:
	case B_USBD_BAD_ARGS:
		ret = B_BAD_VALUE;
		break;

	case B_USBD_NO_MEMORY:
		ret = B_DEV_NO_MEMORY;
		break;

	case B_USBD_PIPE_NOT_CONFIGURED:
		ret = B_DEV_CONFIGURATION_ERROR;
		break;

	case B_USBD_NO_DATA:
	case B_USBD_DEVICE_FAILURE:
	case B_USBD_COMMAND_FAILED:
	case B_USBD_DEVICE_ERROR:
	case B_USBD_PIPE_ERROR:
	default:
		ret = B_IO_ERROR;
		break;
	}
	if (ret != B_OK)
	{
		DPRINTF_ERR ((MY_ID MY_ERR "USBD err %d\n", my_dev->usbd_status));
		return ret;
	}

	/* see USB bus protocol error */

	switch (my_dev->bus_status)
	{
	case B_USB_STATUS_SUCCESS:
		ret = B_OK;
		break;

	case B_USB_STATUS_DEVICE_CRC_ERROR:
		ret = B_IO_ERROR;
		break;

	case B_USB_STATUS_DEVICE_TIMEOUT:
	case B_USB_STATUS_IRP_CANCELLED_BY_REQUEST:
		ret = B_DEV_TIMEOUT;
		break;

	case B_USB_STATUS_DEVICE_STALLED:
		ret = B_IO_ERROR;		/* command error / phase error */
		break;

	case B_USB_STATUS_DRIVER_INTERNAL_ERROR:
	case B_USB_STATUS_ADAPTER_HARDWARE_ERROR:
	case B_USB_STATUS_ISOCH_IRP_ABORTED:
	default:
		ret = B_IO_ERROR;
	}
	if (ret != B_OK)
	{
		DPRINTF_ERR ((MY_ID MY_ERR "USB bus err %d\n", my_dev->bus_status));
		return ret;
	}

	/* see command result */

	if (my_dev->protocol != USB_PROTOCOL_MASS_CB)
	{
		if (my_dev->cmd_status == USB_MASS_STATUS_PASSED)
			return B_OK;
	}
	if (!autosense)
		/* always assume success */
		return B_OK;

	/* reenter: my_dev->xxx_status will be overwritten */
	ret = do_request_sense (my_dev, (void *) &sense, sizeof sense);
	if (ret != B_OK)
		return ret;

	DPRINTF_ERR ((MY_ID MY_ERR "command err ASC:ASCQ=%x:%x\n", sense.asc, sense.ascq));
	switch (sense.asc)
	{
	case 0x00:	ret = B_OK;					break;

	case 0x3a:	ret = B_DEV_NO_MEDIA;		break;
	case 0x04:	ret = B_DEV_NOT_READY;		break;
	case 0x28:	ret = B_DEV_MEDIA_CHANGED;	break;
	case 0x27:	ret = B_READ_ONLY_DEVICE;	break;
	case 0x22:
	case 0x20:	ret = B_DEV_INVALID_IOCTL;	break;

	default:	ret = B_IO_ERROR;			break;
	}
	return ret;
}

/* non-data commands */

/* FIXME: CBI specific */
status_t do_reset (my_device_info *my_dev)
{
	uint8 cdb [MAX_CDB_LEN];
	int cdblen;
	status_t st;

	assert (my_dev != NULL && my_dev->hooks != NULL);
	DPRINTF_INFO ((MY_ID "do_reset()\n"));
	switch (my_dev->protocol)
	{
	case USB_PROTOCOL_MASS_CBI:
	case USB_PROTOCOL_MASS_CB:
		cdblen = build_cdb (cdb, SEND_DIAGNOSTIC, 0, 0, 0, 0);
		cdb [1] = 0x04;
		if (my_dev->cdblen > 0)
			cdblen = my_dev->cdblen;
		(*my_dev->hooks->exec_cmd_nondata)(my_dev, cdb, cdblen);
		st = conv_error (my_dev, false);

		usb->clear_feature (my_dev->bulkInEpt->handle, USB_FEATURE_ENDPOINT_HALT);
		usb->clear_feature (my_dev->bulkOutEpt->handle, USB_FEATURE_ENDPOINT_HALT);
#ifdef DATA_TOGGLE_WORKAROUND
		my_dev->bulkInEptToggle = my_dev->bulkOutEptToggle = 0;
#endif
		break;

	case USB_PROTOCOL_MASS_BULKONLY:
		break;
	}
	return st;
}

status_t do_test_unit_ready (my_device_info *my_dev)
{
	uint8 cdb [MAX_CDB_LEN];
	int cdblen;
	status_t st;

	assert (my_dev != NULL && my_dev->hooks != NULL);
	DPRINTF_INFO ((MY_ID "do_test_unit_ready()\n"));
	cdblen = build_cdb (cdb, TEST_UNIT_READY, 0, 0, 0, 0);
	if (my_dev->cdblen > 0)
		cdblen = my_dev->cdblen;
	(*my_dev->hooks->exec_cmd_nondata)(my_dev, cdb, cdblen);
	st = conv_error (my_dev, true);

	return st;
}

status_t do_prevent_allow_medium_removal (
	my_device_info *my_dev, 
	int flags)
{
	uint8 cdb [MAX_CDB_LEN];
	int cdblen;
	status_t st;

	assert (my_dev != NULL && my_dev->hooks != NULL);
	DPRINTF_INFO ((MY_ID "do_prevent_allow_medium_removal()\n"));
	cdblen = build_cdb (cdb, PREVENT_ALLOW_MEDIUM_REMOVAL, 0, 0, 0, 0);
	cdb [4] = flags;
	if (my_dev->cdblen > 0)
		cdblen = my_dev->cdblen;
	(*my_dev->hooks->exec_cmd_nondata)(my_dev, cdb, cdblen);
	st = conv_error (my_dev, true);

	return st;
}

status_t do_rezero_unit (my_device_info *my_dev)
{
	uint8 cdb [MAX_CDB_LEN];
	int cdblen;
	status_t st;

	assert (my_dev != NULL && my_dev->hooks != NULL);
	DPRINTF_INFO ((MY_ID "do_rezero_unit()\n"));
	cdblen = build_cdb (cdb, REZERO_UNIT, 0, 0, 0, 0);
	if (my_dev->cdblen > 0)
		cdblen = my_dev->cdblen;
	(*my_dev->hooks->exec_cmd_nondata)(my_dev, cdb, cdblen);
	st = conv_error (my_dev, true);

	return st;
}

status_t do_format_unit (my_device_info *my_dev)
{
	uint8 cdb [MAX_CDB_LEN];
	int cdblen;
	status_t st;

	assert (my_dev != NULL && my_dev->hooks != NULL);
	DPRINTF_INFO ((MY_ID "do_format_unit()\n"));
	cdblen = build_cdb (cdb, FORMAT_UNIT, 0, 0, 0, 0);
	if (my_dev->cdblen > 0)
		cdblen = my_dev->cdblen;
	(*my_dev->hooks->exec_cmd_nondata)(my_dev, cdb, cdblen);
	st = conv_error (my_dev, true);

	return st;
}

status_t do_start_stop_unit (
	my_device_info *my_dev, 
	int flags)
{
	uint8 cdb [MAX_CDB_LEN];
	int cdblen;
	status_t st;

	assert (my_dev != NULL && my_dev->hooks != NULL);
	DPRINTF_INFO ((MY_ID "do_start_stop_unit()\n"));
	cdblen = build_cdb (cdb, START_STOP_UNIT, 0, 0, 0, 0);
	cdb [4] = flags;
	if (my_dev->cdblen > 0)
		cdblen = my_dev->cdblen;
	(*my_dev->hooks->exec_cmd_nondata)(my_dev, cdb, cdblen);
	st = conv_error (my_dev, true);

	return st;
}

status_t do_seek_10 (
	my_device_info *my_dev, 
	uint32 lba)
{
	uint8 cdb [MAX_CDB_LEN];
	int cdblen;
	status_t st;

	assert (my_dev != NULL && my_dev->hooks != NULL);
	DPRINTF_INFO ((MY_ID "do_seek_10()\n"));
	cdblen = build_cdb (cdb, SEEK_10, 0, lba, 0, 0);
	if (my_dev->cdblen > 0)
		cdblen = my_dev->cdblen;
	(*my_dev->hooks->exec_cmd_nondata)(my_dev, cdb, cdblen);
	st = conv_error (my_dev, true);

	return st;
}

status_t do_verify_10 (
	my_device_info *my_dev, 
	uint32 lba, 
	uint16 blk)
{
	uint8 cdb [MAX_CDB_LEN];
	int cdblen;
	status_t st;

	assert (my_dev != NULL && my_dev->hooks != NULL);
	DPRINTF_INFO ((MY_ID "do_verify_10()\n"));
	cdblen = build_cdb (cdb, VERIFY_10, 0, lba, blk, 0);
	if (my_dev->cdblen > 0)
		cdblen = my_dev->cdblen;
	(*my_dev->hooks->exec_cmd_nondata)(my_dev, cdb, cdblen);
	st = conv_error (my_dev, true);

	return st;
}

status_t do_synchronize_cache (
	my_device_info *my_dev, 
	uint32 lba, 
	uint16 blk)
{
	uint8 cdb [MAX_CDB_LEN];
	int cdblen;
	status_t st;

	assert (my_dev != NULL && my_dev->hooks != NULL);
	DPRINTF_INFO ((MY_ID "do_synchronize_cache()\n"));
	cdblen = build_cdb (cdb, SYNCHRONIZE_CACHE, 0, lba, blk, 0);
	if (my_dev->cdblen > 0)
		cdblen = my_dev->cdblen;
	(*my_dev->hooks->exec_cmd_nondata)(my_dev, cdb, cdblen);
	st = conv_error (my_dev, true);

	return st;
}

status_t do_play_audio_track_index (
	my_device_info *my_dev, 
	scsi_play_track *track)
{
	uint8 cdb [MAX_CDB_LEN];
	int cdblen;
	status_t st;

	assert (my_dev != NULL && my_dev->hooks != NULL);
	DPRINTF_INFO ((MY_ID "do_play_audio_track_index()\n"));
	cdblen = build_cdb (cdb, PLAY_AUDIO_TRACK_INDEX, 0, 0, 0, 0);
	cdb [4] = track->start_track;
	cdb [5] = track->start_index;
	cdb [7] = track->end_track;
	cdb [8] = track->end_index;
	if (my_dev->cdblen > 0)
		cdblen = my_dev->cdblen;
	(*my_dev->hooks->exec_cmd_nondata)(my_dev, cdb, cdblen);
	st = conv_error (my_dev, true);

	return st;
}

status_t do_play_audio_msf (
	my_device_info *my_dev, 
	scsi_play_position *pos)
{
	uint8 cdb [MAX_CDB_LEN];
	int cdblen;
	status_t st;

	assert (my_dev != NULL && my_dev->hooks != NULL);
	DPRINTF_INFO ((MY_ID "do_play_audio_msf()\n"));
	cdblen = build_cdb (cdb, PLAY_AUDIO_MSF, 0, 0, 0, 0);
	cdb [3] = pos->start_m;
	cdb [4] = pos->start_s;
	cdb [5] = pos->start_f;
	cdb [6] = pos->end_m;
	cdb [7] = pos->end_s;
	cdb [8] = pos->end_f;
	if (my_dev->cdblen > 0)
		cdblen = my_dev->cdblen;
	(*my_dev->hooks->exec_cmd_nondata)(my_dev, cdb, cdblen);
	st = conv_error (my_dev, true);

	return st;
}

status_t do_pause_resume (
	my_device_info *my_dev, 
	int flag)
{
	uint8 cdb [MAX_CDB_LEN];
	int cdblen;
	status_t st;

	assert (my_dev != NULL && my_dev->hooks != NULL);
	DPRINTF_INFO ((MY_ID "do_pause_resume()\n"));
	cdblen = build_cdb (cdb, PAUSE_RESUME, 0, 0, 0, 0);
	cdb [8] = flag;
	if (my_dev->cdblen > 0)
		cdblen = my_dev->cdblen;
	(*my_dev->hooks->exec_cmd_nondata)(my_dev, cdb, cdblen);
	st = conv_error (my_dev, true);

	return st;
}

status_t do_stop_play_scan (my_device_info *my_dev)
{
	uint8 cdb [MAX_CDB_LEN];
	int cdblen;
	status_t st;

	assert (my_dev != NULL && my_dev->hooks != NULL);
	DPRINTF_INFO ((MY_ID "do_stop_play_scan()\n"));
	cdblen = build_cdb (cdb, STOP_PLAY_SCAN, 0, 0, 0, 0);
	if (my_dev->cdblen > 0)
		cdblen = my_dev->cdblen;
	(*my_dev->hooks->exec_cmd_nondata)(my_dev, cdb, cdblen);
	st = conv_error (my_dev, true);

	return st;
}

status_t do_scan_msf (
	my_device_info *my_dev, 
	int dir, 
	scsi_play_position *pos)
{
	uint8 cdb [MAX_CDB_LEN];
	int cdblen;
	status_t st;

	assert (my_dev != NULL && my_dev->hooks != NULL);
	DPRINTF_INFO ((MY_ID "do_scan_msf()\n"));
	cdblen = build_cdb (cdb, SCAN, 0, 0, 0, 0);
	cdb [1] = (dir & 1) << 4;
	cdb [3] = pos->start_m;
	cdb [4] = pos->start_s;
	cdb [5] = pos->start_f;
	cdb [9] = MODE_MSF << 6;	/* MSF format */
	if (my_dev->cdblen > 0)
		cdblen = my_dev->cdblen;
	(*my_dev->hooks->exec_cmd_nondata)(my_dev, cdb, cdblen);
	st = conv_error (my_dev, true);

	return st;
}

status_t do_scan_track (
	my_device_info *my_dev, 
	int dir, 
	int track)
{
	uint8 cdb [MAX_CDB_LEN];
	int cdblen;
	status_t st;

	assert (my_dev != NULL && my_dev->hooks != NULL);
	DPRINTF_INFO ((MY_ID "do_scan_track()\n"));
	cdblen = build_cdb (cdb, SCAN, 0, 0, 0, 0);
	cdb [1] = (dir & 1) << 4;
	cdb [5] = track;
	cdb [9] = MODE_TNO << 6;	/* track number format */
	if (my_dev->cdblen > 0)
		cdblen = my_dev->cdblen;
	(*my_dev->hooks->exec_cmd_nondata)(my_dev, cdb, cdblen);
	st = conv_error (my_dev, true);

	return st;
}

/* data-in commands */

status_t do_inquiry (
	my_device_info *my_dev, 
	uint8 *buf, 
	size_t len)
{
	uint8 cdb [MAX_CDB_LEN];
	int cdblen;
	status_t st;

	assert (my_dev != NULL && my_dev->hooks != NULL && buf != NULL);
	DPRINTF_INFO ((MY_ID "inquiry()\n"));
	if (len > INQUIRY_LEN)
		len = INQUIRY_LEN;
	cdblen = build_cdb (cdb, INQUIRY, 0, 0, len, 0);
	if (my_dev->cdblen > 0)
		cdblen = my_dev->cdblen;
	(*my_dev->hooks->exec_cmd_datain)(my_dev, cdb, cdblen, buf, len);
	st = conv_error (my_dev, false);

	return st;
}

status_t do_request_sense (
	my_device_info *my_dev, 
	uint8 *buf, 
	size_t len)
{
	uint8 cdb [MAX_CDB_LEN];
	int cdblen;
	status_t st;

	assert (my_dev != NULL && my_dev->hooks != NULL && buf != NULL);
	DPRINTF_INFO ((MY_ID "request_sense()\n"));
	if (len > SENSE_LEN)
		len = SENSE_LEN;
	cdblen = build_cdb (cdb, REQUEST_SENSE, 0, 0, len, 0);
	if (my_dev->cdblen > 0)
		cdblen = my_dev->cdblen;
	(*my_dev->hooks->exec_cmd_datain)(my_dev, cdb, cdblen, buf, len);
	st = conv_error (my_dev, false);

	return st;
}

status_t do_mode_sense_10 (
	my_device_info *my_dev, 
	uint8 *buf, 
	size_t len, 
	int pc, 
	int page)
{
	uint8 cdb [MAX_CDB_LEN];
	int cdblen;
	status_t st;

	assert (my_dev != NULL && my_dev->hooks != NULL && buf != NULL);
	DPRINTF_INFO ((MY_ID "mode_sense()\n"));
	cdblen = build_cdb (cdb, MODE_SENSE_10, 0, 0, len, 0);
	cdb [2] = (pc << 6) | page;
	if (my_dev->cdblen > 0)
		cdblen = my_dev->cdblen;
	(*my_dev->hooks->exec_cmd_datain)(my_dev, cdb, cdblen, buf, len);
	st = conv_error (my_dev, true);

	return st;
}

status_t do_read_10 (
	my_device_info *my_dev, 
	const iovec *vec, 
	size_t vecsiz, 
	uint32 lba, 
	uint16 blk)
{
	uint8 cdb [MAX_CDB_LEN];
	int cdblen;
	status_t st;

	assert (my_dev != NULL && my_dev->hooks != NULL && vec != NULL);
	DPRINTF_INFO ((MY_ID "do_read_10(%d,%u,%u)\n", my_dev->number, lba, blk));
	cdblen = build_cdb (cdb, READ_10, 0, lba, blk, 0);
	if (my_dev->cdblen > 0)
		cdblen = my_dev->cdblen;
	(*my_dev->hooks->exec_cmd_datain_sg)(
		my_dev, cdb, cdblen, vec, vecsiz, my_dev->sector_size * blk);
	st = conv_error (my_dev, true);

	return st;
}

status_t do_read_12 (
	my_device_info *my_dev, 
	const iovec *vec, 
	size_t vecsiz, 
	uint32 lba, 
	uint32 blk)
{
	uint8 cdb [MAX_CDB_LEN];
	int cdblen;
	status_t st;

	assert (my_dev != NULL && my_dev->hooks != NULL && vec != NULL);
	DPRINTF_INFO ((MY_ID "do_read_12()\n"));
	cdblen = build_cdb (cdb, READ_12, 0, lba, blk, 0);
	if (my_dev->cdblen > 0)
		cdblen = my_dev->cdblen;
	(*my_dev->hooks->exec_cmd_datain_sg)
		(my_dev, cdb, cdblen, vec, vecsiz, my_dev->sector_size * blk);
	st = conv_error (my_dev, true);

	return st;
}

status_t do_read_capacity (
	my_device_info *my_dev, 
	uint8 *buf, 
	size_t len)
{
	uint8 cdb [MAX_CDB_LEN];
	int cdblen;
	status_t st;

	assert (my_dev != NULL && my_dev->hooks != NULL && buf != NULL);
	DPRINTF_INFO ((MY_ID "do_read_capacity()\n"));
	cdblen = build_cdb (cdb, READ_CAPACITY, 0, 0, 0, 0);
	if (my_dev->cdblen > 0)
		cdblen = my_dev->cdblen;
	(*my_dev->hooks->exec_cmd_datain)(my_dev, cdb, cdblen, buf, len);
	st = conv_error (my_dev, true);

	return st;
}

/* XXX UFI specific */
status_t do_read_format_capacities (
	my_device_info *my_dev, 
	uint8 *buf, 
	size_t len)
{
	uint8 cdb [MAX_CDB_LEN];
	int cdblen;
	status_t st;

	assert (my_dev != NULL && my_dev->hooks != NULL && buf != NULL);
	DPRINTF_INFO ((MY_ID "do_read_format_capacities()\n"));
	cdblen = build_cdb (cdb, READ_FORMAT_CAPACITIES, 0, 0, len, 0);
	if (my_dev->cdblen > 0)
		cdblen = my_dev->cdblen;
	(*my_dev->hooks->exec_cmd_datain)(my_dev, cdb, cdblen, buf, len);
	st = conv_error (my_dev, true);

	return st;
}

status_t do_read_toc (
	my_device_info *my_dev, 
	uint8 *buf, 
	size_t len, 
	int mode, 
	int track)
{
	uint8 cdb [MAX_CDB_LEN];
	int cdblen;
	status_t st;

	assert (my_dev != NULL && my_dev->hooks != NULL && buf != NULL);
	DPRINTF_INFO ((MY_ID "do_read_toc()\n"));
	cdblen = build_cdb (cdb, READ_TOC, 0, 0, len, 0);
	if (mode == MODE_MSF)
		cdb [1] |= 2;
	cdb [6] = track;
	if (my_dev->cdblen > 0)
		cdblen = my_dev->cdblen;
	(*my_dev->hooks->exec_cmd_datain)(my_dev, cdb, cdblen, buf, len);
	st = conv_error (my_dev, true);

	return st;
}

status_t do_read_sub_channel (
	my_device_info *my_dev, 
	uint8 *buf, 
	size_t len, 
	int mode, 
	int track, 
	int format)
{
	uint8 cdb [MAX_CDB_LEN];
	int cdblen;
	status_t st;

	assert (my_dev != NULL && my_dev->hooks != NULL && buf != NULL);
	DPRINTF_INFO ((MY_ID "do_read_sub_channel()\n"));
	cdblen = build_cdb (cdb, READ_SUB_CHANNEL, 0, 0, len, 0);
	if (mode == MODE_MSF)
		cdb [1] |= 2;
	cdb [2] = 0x40;		/* set SubQ */
	cdb [6] = track;
	cdb [3] = format;
	if (my_dev->cdblen > 0)
		cdblen = my_dev->cdblen;
	(*my_dev->hooks->exec_cmd_datain)(my_dev, cdb, cdblen, buf, len);
	st = conv_error (my_dev, true);

	return st;
}

status_t do_read_cd_msf (
	my_device_info *my_dev, 
	uint8 *buf, 
	size_t len, 
	scsi_play_position *pos)
{
	uint8 cdb [MAX_CDB_LEN];
	int cdblen;
	status_t st;

	assert (my_dev != NULL && my_dev->hooks != NULL && buf != NULL);
	DPRINTF_INFO ((MY_ID "do_read_cd_msf()\n"));
	cdblen = build_cdb (cdb, READ_CD_MSF, 0, 0, 0, 0);
	cdb [1] = 1 << 2;	/* CD-DA expected */
	cdb [3] = pos->start_m;
	cdb [4] = pos->start_s;
	cdb [5] = pos->start_f;
	cdb [6] = pos->end_m;
	cdb [7] = pos->end_s;
	cdb [8] = pos->end_f;
	cdb [9] = 0x10;		/* user data only */
	if (my_dev->cdblen > 0)
		cdblen = my_dev->cdblen;
	(*my_dev->hooks->exec_cmd_datain)(my_dev, cdb, cdblen, buf, len);
	st = conv_error (my_dev, true);

	return st;
}

/* data-out commands */

status_t do_write_10 (
	my_device_info *my_dev, 
	const iovec *vec, 
	size_t vecsiz, 
	uint32 lba, 
	uint16 blk)
{
	uint8 cdb [MAX_CDB_LEN];
	int cdblen;
	status_t st;

	assert (my_dev != NULL && my_dev->hooks != NULL && vec != NULL);
	DPRINTF_INFO ((MY_ID "do_write_10()\n"));
	cdblen = build_cdb (cdb, WRITE_10, 0, lba, blk, 0);
	if (my_dev->cdblen > 0)
		cdblen = my_dev->cdblen;
	(*my_dev->hooks->exec_cmd_dataout_sg)(my_dev, cdb, cdblen, vec, vecsiz, my_dev->sector_size * blk);
	st = conv_error (my_dev, true);

	return st;
}

status_t do_write_12 (
	my_device_info *my_dev, 
	const iovec *vec, 
	size_t vecsiz, 
	uint32 lba, 
	uint32 blk)
{
	uint8 cdb [MAX_CDB_LEN];
	int cdblen;
	status_t st;

	assert (my_dev != NULL && my_dev->hooks != NULL && vec != NULL);
	DPRINTF_INFO ((MY_ID "do_write_12()\n"));
	cdblen = build_cdb (cdb, WRITE_12, 0, lba, blk, 0);
	if (my_dev->cdblen > 0)
		cdblen = my_dev->cdblen;
	(*my_dev->hooks->exec_cmd_dataout_sg)
		(my_dev, cdb, cdblen, vec, vecsiz, my_dev->sector_size * blk);
	st = conv_error (my_dev, true);

	return st;
}

status_t do_write_and_verify_10 (
	my_device_info *my_dev, 
	const iovec *vec, 
	size_t vecsiz, 
	uint32 lba, 
	uint16 blk)
{
	uint8 cdb [MAX_CDB_LEN];
	int cdblen;
	status_t st;

	assert (my_dev != NULL && my_dev->hooks != NULL && vec != NULL);
	DPRINTF_INFO ((MY_ID "do_write_and_verify()\n"));
	cdblen = build_cdb (cdb, WRITE_AND_VERIFY_10, 0, lba, blk, 0);
	if (my_dev->cdblen > 0)
		cdblen = my_dev->cdblen;
	(*my_dev->hooks->exec_cmd_dataout_sg)
		(my_dev, cdb, cdblen, vec, vecsiz, my_dev->sector_size * blk);
	st = conv_error (my_dev, true);

	return st;
}

/* XXX UFI specific extension */
status_t do_format_unit_ufi (
	my_device_info *my_dev, 
	const uint8 *buf, 
	size_t len, 
	int track)
{
	uint8 cdb [MAX_CDB_LEN];
	status_t st;

	assert (my_dev != NULL && my_dev->hooks != NULL && buf != NULL);
	DPRINTF_INFO ((MY_ID "do_format_unit()\n"));
	/* 10-byte command format in spite of Group 0 */
	memset (cdb, 0, MAX_CDB_LEN);
	cdb [0] = FORMAT_UNIT;
	cdb [1] = 0x17;
	cdb [2] = track;
	cdb [7] = len >> 8;
	cdb [8] = len;
	(*my_dev->hooks->exec_cmd_dataout)
		(my_dev, cdb, my_dev->cdblen, buf, len);
	st = conv_error (my_dev, true);

	return st;
}

status_t do_mode_select_10 (
	my_device_info *my_dev, 
	const uint8 *buf, 
	size_t len)
{
	uint8 cdb [MAX_CDB_LEN];
	int cdblen;
	status_t st;

	assert (my_dev != NULL && my_dev->hooks != NULL && buf != NULL);
	DPRINTF_INFO ((MY_ID "do_mode_select_10()\n"));
	cdblen = build_cdb (cdb, MODE_SELECT_10, 0, 0, len, 0);
	cdb [1] |= 0x10;
	if (my_dev->cdblen > 0)
		cdblen = my_dev->cdblen;
	(*my_dev->hooks->exec_cmd_dataout)
		(my_dev, cdb, cdblen, buf, MODE_HDR_LEN);
	st = conv_error (my_dev, true);

	return st;
}

/* finialize (workaround) */
/* XXX assuming no error */

#ifdef DATA_TOGGLE_WORKAROUND
status_t do_finish (my_device_info *my_dev)
{
	uint8	cdb [MAX_CDB_LEN], buf [64];
	int cdblen;

	assert (my_dev != NULL && my_dev->hooks != NULL);
	DPRINTF_INFO ((MY_ID "do_finish()\n"));

	switch (my_dev->protocol)
	{
	case USB_PROTOCOL_MASS_CBI:
	case USB_PROTOCOL_MASS_CB:
		if (my_dev->bulkOutEptToggle)
		{
			/* data-out command (MODE SELECT) */
			cdblen = build_cdb (cdb, MODE_SELECT_10, 0, 0, MODE_HDR_LEN, 0);
			cdb [1] |= 0x10;
			memset (buf, 0, MODE_HDR_LEN);
			if (my_dev->cdblen > 0)
				cdblen = my_dev->cdblen;
			(*my_dev->hooks->exec_cmd_dataout)
				(my_dev, cdb, cdblen, buf, MODE_HDR_LEN);
		}
		if (my_dev->bulkInEptToggle)
		{
			/* data-in command (INQUIRY) */
			cdblen = build_cdb (cdb, INQUIRY, 0, 0, 0, 0);
			cdb [4] = 8;		/* only one bulk packet */
			if (my_dev->cdblen > 0)
				cdblen = my_dev->cdblen;
			(*my_dev->hooks->exec_cmd_datain)(my_dev, cdb, cdblen, buf, 8);
		}
		if (my_dev->intrEptToggle)
		{
			/* non-data command (INQUIRY) */
			cdblen = build_cdb (cdb, INQUIRY, 0, 0, 0, 0);
			cdb [4] = 0;
			if (my_dev->cdblen > 0)
				cdblen = my_dev->cdblen;
			(*my_dev->hooks->exec_cmd_nondata)(my_dev, cdb, cdblen);
		}
		break;

	case USB_PROTOCOL_MASS_BULKONLY:
		/* XXX assuming CBW and CSW fit in a bulk packet */
		switch ((my_dev->bulkInEptToggle << 1) | my_dev->bulkOutEptToggle)
		{
		case 0:
			/* nop */
			break;

		case 1:
			/* data-in command (INQUIRY) */
			cdblen = build_cdb (cdb, INQUIRY, 0, 0, 0, 0);
			cdb [4] = 8;		/* only one bulk packet */
			if (my_dev->cdblen > 0)
				cdblen = my_dev->cdblen;
			(*my_dev->hooks->exec_cmd_datain)(my_dev, cdb, cdblen, buf, 8);
			break;

		case 2:
			/* data-out command (MODE SELECT) */
			cdblen = build_cdb (cdb, MODE_SELECT_10, 0, 0, MODE_HDR_LEN, 0);
			cdb [1] |= 0x10;
			memset (buf, 0, MODE_HDR_LEN);
			if (my_dev->cdblen > 0)
				cdblen = my_dev->cdblen;
			(*my_dev->hooks->exec_cmd_dataout)
				(my_dev, cdb, cdblen, buf, MODE_HDR_LEN);
			break;

		case 3:
			/* non-data command (INQUIRY) */
			cdblen = build_cdb (cdb, INQUIRY, 0, 0, 0, 0);
			cdb [4] = 0;
			if (my_dev->cdblen > 0)
				cdblen = my_dev->cdblen;
			(*my_dev->hooks->exec_cmd_nondata)(my_dev, cdb, cdblen);
			break;
		}
		break;
	}
	assert (my_dev->bulkInEptToggle == 0 && 
		my_dev->bulkOutEptToggle == 0 &&
		my_dev->intrEptToggle == 0);
	return B_OK;
}
#endif
