/*
	USB mass storage driver for BeOS R5
	Copyright (C) 2000 ITO, Takayuki
	All rights reserved
*/

#include "driver.h"

#define	DPRINTF_INFO(x)	dprintf x
#define	DPRINTF_ERR(x)	dprintf x

typedef struct my_open_info
{
	struct my_open_info *next;
	my_device_info	*my_dev;
	int status;
} my_open_info;

usb_module_info *usb;

const char *my_driver_name = "usb_mass";
const char *my_base_name = "disk/usb/";

/*
	publish_devices
*/

const char **publish_devices (void)
{
	DPRINTF_INFO ((MY_ID "publish_devices()\n"));

	if (my_device_list_changed)
	{
		free_device_names ();
		alloc_device_names ();
		if (my_device_names != NULL)
			rebuild_device_names ();
		my_device_list_changed = false;
	}
	assert (my_device_names != NULL);
	return (const char **) my_device_names;
}

/*
	unusual devices
*/

unusual_dev unusual_dev_list [] =
{
	// HP USB CD-Writer Plus
	{ 0x03f0, 0x0107, 0x0200, 
		USB_SUBCLASS_MASS_SFF8070, USB_PROTOCOL_MASS_CBI, 0 },
	// Matsushita LS-120, Imation SuperDisk (Shuttle E-USB Bridge)
	{ 0x04e6, 0x0001, 0x0200,
		USB_SUBCLASS_MASS_SFF8020, USB_PROTOCOL_MASS_CBI, 0 },
	// Shuttle eUSCSI Bridge
	{ 0x04e6, 0x0002, 0x0100,
		USB_SUBCLASS_MASS_SCSI, USB_PROTOCOL_MASS_BULKONLY, 0 }, 
	// Shuttle eUSB MMC Adapter
	{ 0x04e6, 0x0006, 0x0100,
		USB_SUBCLASS_MASS_SCSI, USB_PROTOCOL_MASS_CBI, 0 }, 
	// Y-E Data FlashBuster-U (old rev.)
	{ 0x057b, 0x0000, 0x0114,
		USB_SUBCLASS_MASS_UFI, USB_PROTOCOL_MASS_CBI, 0 },
	// Iomega zip 250
	{ 0x059b, 0x0030, 0x0100,
		USB_SUBCLASS_MASS_SCSI, USB_PROTOCOL_MASS_BULKONLY, 0 },
	// Hagiwara Sys-Com FlashGate SmartMedia
	{ 0x0693, 0x0002, 0x0100,
		USB_SUBCLASS_MASS_SCSI, USB_PROTOCOL_MASS_BULKONLY, 0 },
	// SanDisk ImageMate (SDDR-01,05)
	{ 0x0781, 0x0001, 0x0200,
		USB_SUBCLASS_MASS_SCSI, USB_PROTOCOL_MASS_CBI, 0 },
	// SanDisk ImageMate (SDDR-31)
	{ 0x0781, 0x0002, 0x0009,
		USB_SUBCLASS_MASS_SCSI, USB_PROTOCOL_MASS_BULKONLY, 0 },
	// Microtech USB-SCSI-HD50
	{ 0x07af, 0x0005, 0x0100,
		USB_SUBCLASS_MASS_SCSI, USB_PROTOCOL_MASS_BULKONLY, 0 }, 
	{ 0x0000, 0x0000, 0x0000, 0, 0, 0 }
};

/*
	USB specific device hooks
*/

static status_t my_device_added (
	const usb_device *dev, 
	void **cookie)
{
	my_device_info *my_dev;
	const usb_configuration_info *conf;
	const usb_interface_info *ii;
	status_t st;
	sense_data sense;
	const usb_device_descriptor *ddesc;
	const unusual_dev *un_dev;

	assert (dev != NULL && cookie != NULL);
	DPRINTF_INFO ((MY_ID "device_added()\n"));

	if ((conf = usb->get_nth_configuration 
		(dev, DEFAULT_CONFIGURATION - 1)) == NULL)
	{
		DPRINTF_ERR ((MY_ID "cannot get default configuration\n"));
		return B_ERROR;
	}

	ddesc = usb->get_device_descriptor (dev);
	for (un_dev = unusual_dev_list; un_dev->vendor_id != 0; un_dev++)
	{
		if (ddesc->vendor_id == un_dev->vendor_id &&
			ddesc->product_id == un_dev->product_id &&
			ddesc->device_version == un_dev->device_version)
		{
			DPRINTF_INFO((MY_ID "unusual device found\n"));
			break;
		}
	}
	ii = conf->interface [DEFAULT_INTERFACE - 1].active;
	if (un_dev->vendor_id == 0)
	{
		/* not listed */
		un_dev = NULL;
		if (ii->descr->interface_class != USB_CLASS_MASS)
		{
			DPRINTF_INFO ((MY_ID "not mass storage\n"));
			return B_ERROR;
		}
	}

	if ((st = usb->set_configuration (dev, conf)) != B_OK)
	{
		DPRINTF_ERR ((MY_ID "set_configuration() failed %d\n", (int)st));
		return B_ERROR;
	}

	if ((my_dev = create_device (dev, ii, un_dev)) == NULL)
		return B_ERROR;

	add_device_info (my_dev);
	*cookie = my_dev;
	DPRINTF_INFO ((MY_ID "added %d\n", my_dev->number));

/*	acquire_sem (my_dev->sem_lock); */
#if 1
	do_reset (my_dev);
#else
	/* to clear Unit Attention: ASC = 0x29 */
	do_request_sense (my_dev, (void *) &sense, sizeof sense);
#endif
	do_inquiry (my_dev, (void *) &my_dev->inq, sizeof (std_inquiry_data));
	DPRINTF_INFO ((MY_ID "%-.8s %-.16s %-.4s\n", my_dev->inq.vendor_id, 
		my_dev->inq.product_id, my_dev->inq.product_version));
	do_test_unit_ready (my_dev);
#ifdef DATA_TOGGLE_WORKAROUND
	do_finish (my_dev);
#endif
/*	release_sem (my_dev->sem_lock); */

	return B_OK;
}

static status_t my_device_removed (void *cookie)
{
	my_device_info *my_dev = cookie;

	assert (cookie != NULL);

	DPRINTF_INFO ((MY_ID "device_removed(%d)\n", my_dev->number));
	remove_device_info (my_dev);
	if (my_dev->open == 0)
		remove_device (my_dev);
	else
	{
		DPRINTF_INFO ((MY_ID "%s%d still open\n", my_base_name, my_dev->number));
		my_dev->active = false;
	}
	return B_OK;
}

static usb_notify_hooks my_notify_hooks =
{
	my_device_added, my_device_removed
};

#define	SUPPORTED_DEVICES	3
usb_support_descriptor my_supported_devices [SUPPORTED_DEVICES] =
{
	{ USB_CLASS_MASS,   0, 0, 0, 0 },
	{ USB_CLASS_VENDOR, 0, 0, 0, 0 },
	{ USB_CLASS_UNSPEC, 0, 0, 0, 0 }
};

/*
	init_hardware
*/

status_t init_hardware (void)
{
	DPRINTF_INFO ((MY_ID "init_hardware() " __DATE__ " " __TIME__"\n"));
	return B_OK;
}


/* 
	init_driver
*/

status_t init_driver (void)
{
	DPRINTF_INFO ((MY_ID "init_driver() " __DATE__ " " __TIME__"\n"));

/*	load_driver_symbols (my_driver_name); */

	if (get_module (B_USB_MODULE_NAME, (module_info **) &usb) != B_OK)
		return B_ERROR;

	if ((my_device_list_lock = create_sem (1, "dev_list_lock")) < 0)
	{
		put_module (B_USB_MODULE_NAME);
		return my_device_list_lock;		/* error code */
	}

	usb->register_driver (my_driver_name, my_supported_devices, 
		SUPPORTED_DEVICES, NULL);
	usb->install_notify (my_driver_name, &my_notify_hooks);
	DPRINTF_INFO ((MY_ID "init_driver() OK\n"));
	return B_OK;
}


/*
	uninit_driver
*/

void uninit_driver (void)
{
	DPRINTF_INFO ((MY_ID "uninit_driver()\n"));

	usb->uninstall_notify (my_driver_name);
	delete_sem (my_device_list_lock);
	put_module (B_USB_MODULE_NAME);
	free_device_names ();
}

/*
	my_device_open
*/

static status_t my_device_open (
	const char *name, 
	uint32 flags, 
	void **cookie)
{
	my_device_info *my_dev;
	my_open_info *my_fd;
	int number, lun, ret = B_ERROR;
	char pttn [16];	/* XXX */
	const char *p;

	assert (name != NULL && cookie != NULL);
	DPRINTF_INFO ((MY_ID "open(%s)\n", name));

	if (strlen (name) < strlen (my_base_name) + 7)	/* "0/0/raw" */
		return B_ENTRY_NOT_FOUND;
	p = &name [strlen (my_base_name)];
	number = atoi (p);
	if ((p = strchr (p, '/')) == NULL)
		return B_ENTRY_NOT_FOUND;
	lun = atoi (++p);
	if ((p = strchr (p, '/')) == NULL)
		return B_ENTRY_NOT_FOUND;
	strcpy (pttn, ++p);

	if ((my_dev = search_device_info (number)) == NULL)
		ret = B_ENTRY_NOT_FOUND;
	else if ((my_fd = malloc (sizeof (my_open_info))) == NULL)
		ret = B_NO_MEMORY;
	else
	{
		acquire_sem (my_dev->sem_lock);
		my_fd->status = B_OK;
		my_fd->my_dev = my_dev;
		my_fd->next = my_dev->open_fds;
		my_dev->open_fds = my_fd;
		my_dev->open++;
		release_sem (my_dev->sem_lock);
		*cookie = my_fd;
		DPRINTF_INFO ((MY_ID "device %d open (%d)\n", number, my_dev->open));
		ret = B_OK;
	}
	return ret;
}


/*
	my_device_read
*/

static status_t my_device_readv (
	void *cookie, 
	off_t position, 
	const iovec *vec, size_t count, 
	size_t *num_bytes)
{
	my_open_info *my_fd = cookie;
	my_device_info *my_dev;
	status_t st;
	int i, sector_size;

	assert (cookie != NULL && vec != NULL && num_bytes != NULL);
	my_dev = my_fd->my_dev;
	sector_size = my_dev->sector_size;
	DPRINTF_INFO ((MY_ID "readv(%d,%d,%d), %d bytes/sector\n", 
		my_dev->number, (int)(position / sector_size), 
		(int)count, (int)sector_size));

	if (my_fd->status == B_DEV_MEDIA_CHANGED)
	{
		DPRINTF_ERR ((MY_ID MY_ERR "medium changed\n"));
		return B_DEV_MEDIA_CHANGED;
	}
	st = do_test_unit_ready (my_dev);
	if (st != B_OK)
	{
		if (st == B_DEV_MEDIA_CHANGED)
		{
			/* propagate to all open fds */
			my_open_info *p;
			for (p = my_dev->open_fds; p != NULL; p = p->next)
				p->status = B_DEV_MEDIA_CHANGED;
			DPRINTF_ERR ((MY_ID MY_ERR "medium changed\n"));
		}
		return st;
	}
	
	if (position & (sector_size - 1))
		DPRINTF_ERR ((MY_ID MY_WARN "not sector boundary\n"));
	/* FIXME */
	if (count == 1)
	{
		if (vec [0].iov_len & (sector_size - 1))
			DPRINTF_ERR ((MY_ID MY_WARN "%d not sector size\n", 
				(int)vec [0].iov_len));
	}
	else
	{
		if (count > 64)
			DPRINTF_ERR ((MY_ID MY_WARN "too many iovec entries\n"));
		for (i = 0; i < count; i++)
			if (vec [i].iov_len != sector_size)
				DPRINTF_ERR ((MY_ID MY_WARN "not sector size\n"));
	}

/*	acquire_sem (my_dev->sem_lock); */
	st = do_read_10 (my_dev, vec, count, position / sector_size, 
		(count == 1) ? ((vec[0].iov_len + sector_size - 1) / sector_size) 
			: count);	/* FIXME */
	*num_bytes = my_dev->bulk_length;
#ifdef DATA_TOGGLE_WORKAROUND
	do_finish (my_dev);
#endif
/*	release_sem (my_dev->sem_lock); */
	if (st < 0)
		*num_bytes = 0;
	return st;
}

static status_t my_device_read (
	void *cookie, 
	off_t position, 
	void *buf, size_t *num_bytes)
{
	iovec vec;
	vec.iov_base = buf;
	vec.iov_len = *num_bytes;
	return my_device_readv (cookie, position, &vec, 1, num_bytes);
}


/*
	my_device_write
*/

static status_t my_device_writev (
	void *cookie, 
	off_t position, 
	const iovec *vec, size_t count, 
	size_t *num_bytes)
{
	my_open_info *my_fd = cookie;
	my_device_info *my_dev;
	status_t st;
	int i, sector_size;

	assert (cookie != NULL && vec != NULL && num_bytes != NULL);
	my_dev = my_fd->my_dev;
	sector_size = my_dev->sector_size;
	DPRINTF_INFO ((MY_ID "writev(%d,%d,%d), %d bytes/sector\n", 
		my_dev->number, (int)(position / sector_size), 
		(int)count, (int)sector_size));

	if (my_fd->status == B_DEV_MEDIA_CHANGED)
	{
		DPRINTF_ERR ((MY_ID MY_ERR "medium changed\n"));
		return B_DEV_MEDIA_CHANGED;
	}
	st = do_test_unit_ready (my_dev);
	if (st != B_OK)
	{
		if (st == B_DEV_MEDIA_CHANGED)
		{
			/* propagate to all open fds */
			my_open_info *p;
			for (p = my_dev->open_fds; p != NULL; p = p->next)
				p->status = B_DEV_MEDIA_CHANGED;
			DPRINTF_ERR ((MY_ID MY_ERR "medium changed\n"));
		}
		return st;
	}

	{
		mode_param_header_8 header;

		st = do_mode_sense_10 (my_dev, (void *) &header, sizeof header, 
			0, 0x3f);
		if (st == B_OK && header.wp)
			return B_READ_ONLY_DEVICE;
	}
	
	if (position & (sector_size - 1))
		DPRINTF_ERR ((MY_ID MY_WARN "not sector boundary\n"));
	/* FIXME */
	if (count == 1)
	{
		if (vec [0].iov_len & (sector_size - 1))
			DPRINTF_ERR ((MY_ID MY_WARN "%d not sector size\n", 
				(int)vec [0].iov_len));
	}
	else
	{
		if (count > 64)
			DPRINTF_ERR ((MY_ID MY_WARN "too many iovec entries\n"));
		for (i = 0; i < count; i++)
			if (vec [i].iov_len != sector_size)
				DPRINTF_ERR ((MY_ID MY_WARN "not sector size\n"));
	}

/*	acquire_sem (my_dev->sem_lock); */
	st = do_write_10 (my_dev, vec, count, position / sector_size, 
		(count == 1) ? ((vec[0].iov_len + sector_size - 1) / sector_size) 
			: count);	/* FIXME */
	*num_bytes = my_dev->bulk_length;
#ifdef DATA_TOGGLE_WORKAROUND
	do_finish (my_dev);
#endif
/*	release_sem (my_dev->sem_lock); */
	if (st < 0)
		*num_bytes = 0;
	return st;
}

static status_t my_device_write (
	void *cookie, 
	off_t position, 
	const void *buf, size_t *num_bytes)
{
	iovec vec;
	vec.iov_base = (void *) buf;
	vec.iov_len = *num_bytes;
	return my_device_writev (cookie, position, &vec, 1, num_bytes);
}

/*
	my_device_control
*/

static status_t get_geometry (
	my_device_info *my_dev, 
	device_geometry *dinfo)
{
	uint8 buf [256];
	status_t ret;
	read_capacity_data cap;

	memset (dinfo, 0, sizeof (device_geometry));
	dinfo->device_type = my_dev->inq.device_type;
		/* B_DISK == B_SCSI_DISK etc. */
	dinfo->removable = my_dev->inq.removable;
	dinfo->write_once = dinfo->device_type == B_WORM;
	dinfo->read_only = true;	/* FIXME */

	ret = do_test_unit_ready (my_dev);
	if (ret != B_OK)
		return ret;

	/* is it something like a floppy disk? */
	ret = do_mode_sense_10 (my_dev, buf, sizeof (mode_param_header_8) + 
		sizeof (flexible_disk_page), 0, 5);
	if (ret == B_OK)
	{
		flexible_disk_page *fd_page;
		fd_page = (flexible_disk_page *) &buf [sizeof (mode_param_header_8)];
		dinfo->bytes_per_sector  = read2 (fd_page->sector_size);
		dinfo->sectors_per_track = fd_page->num_sectors_per_track;
		dinfo->cylinder_count    = read2 (fd_page->num_cylinders);
		dinfo->head_count        = fd_page->num_heads;
		return ret;
	}

	ret = do_read_capacity (my_dev, (void *)&cap, sizeof cap);
	if (ret != B_OK)
		return ret;
	dinfo->bytes_per_sector  = read4 (cap.block_length);
	dinfo->cylinder_count = read4 (cap.last_lba) + 1;
	dinfo->head_count = dinfo->sectors_per_track = 1;

#if 0
	/* is it something like a hard disk? */
	ret  =do_mode_sense_10 (my_dev, buf, sizeof (mode_param_header_8) + 
		sizeof (fixed_drive_page), 0, 5);
	if (ret == B_OK)
	{
		fixed_drive_page *hd_page;
		hd_page = (fixed_drive_page *) &buf [sizeof (mode_param_header_8)];
		dinfo->cylinder_count    = read3 (hd_page->num_cylinders);
		dinfo->head_count        = hd_page->num_heads;
		dinfo->sectors_per_track = (read4 (cap.last_lba) + 1) /
			dinfo->cylinder_count / dinfo->head_count;
		return ret;
	}
#endif

	return ret;
}

static status_t get_icon (
	my_device_info *my_dev, 
	device_icon *icon)
{
	status_t ret;
	const uchar *largeIcon, *miniIcon;
	int device_type;

	device_type = my_dev->inq.device_type;
	if (device_type == B_SCSI_CD || device_type == B_SCSI_RMC_CD)
	{
		largeIcon = largeCDIcon;
		miniIcon = miniCDIcon;
	}
	else if (my_dev->inq.removable && 
		(device_type == B_SCSI_DISK || device_type == B_SCSI_OPTICAL || 
		 device_type == B_SCSI_WORM || device_type == B_SCSI_RBC_DISK))
	{
		if (my_dev->subclass == USB_SUBCLASS_MASS_UFI)
		{
			/* probably FDD */
			largeIcon = largeFloppyIcon;
			miniIcon = miniFloppyIcon;
		}
		else
		{
			largeIcon = largeRemovableIcon;
			miniIcon = miniRemovableIcon;
		}
	}
	else
	{
		largeIcon = largeDiskIcon;
		miniIcon = miniDiskIcon;
	}

	if (icon->icon_size == B_LARGE_ICON)
	{
		memcpy (icon->icon_data, largeIcon, B_LARGE_ICON * B_LARGE_ICON);
		ret = B_OK;
	}
	else if (icon->icon_size == B_MINI_ICON)
	{
		memcpy (icon->icon_data, miniIcon, B_MINI_ICON * B_MINI_ICON);
		ret = B_OK;
	}
	else
	{
		DPRINTF_ERR ((MY_ID MY_ERR "unknown icon size %d\n", 
			(int)icon->icon_size));
		ret = B_BAD_TYPE;
	}
	return ret;
}

static status_t my_device_control (
	void *cookie, 
	uint32 op, 
	void *arg, size_t len)
{
	my_open_info *my_fd = cookie;
	my_device_info *my_dev;
	int number;
	status_t ret = B_DEV_INVALID_IOCTL, st;

	assert (cookie != NULL && my_fd->my_dev != NULL);
	my_dev = my_fd->my_dev;
	number = my_dev->number;
	switch (op)
	{
	case B_GET_GEOMETRY:
		/* very important */
		DPRINTF_INFO ((MY_ID "ioctl(%d, B_GET_GEOMETRY)\n", number));
		{
			device_geometry *dinfo;
			dinfo = (device_geometry *) arg;
			ret = get_geometry (my_dev, dinfo);
			if (ret == B_OK)
			{
				my_dev->sector_size = dinfo->bytes_per_sector;
				DPRINTF_INFO (("%u cylinders, %u heads, %u sectors, %u bytes\n",
					(uint)dinfo->cylinder_count, 
					(uint)dinfo->head_count, 
					(uint)dinfo->sectors_per_track, 
					(uint)dinfo->bytes_per_sector));
			}
			else
				DPRINTF_ERR ((MY_ID MY_ERR "failed to get geometry\n"));
		}
		break;
		
	case B_GET_MEDIA_STATUS:
		/* very important */
		DPRINTF_INFO ((MY_ID "ioctl(%d, B_GET_MEDIA_STATUS)\n", number));
		st = do_test_unit_ready (my_dev);
		if (st == B_DEV_MEDIA_CHANGED)
		{
			/* propagate to all open fds */
			my_open_info *p;
			for (p = my_dev->open_fds; p != NULL; p = p->next)
				p->status = B_DEV_MEDIA_CHANGED;
			DPRINTF_INFO ((MY_ID "medium changed\n"));
		}
		else if (my_fd->status == B_DEV_MEDIA_CHANGED)
		{
			st = B_DEV_MEDIA_CHANGED;
		}
		*(status_t *) arg = st;
		my_fd->status = B_OK;
		ret = B_OK;		/* FIXME */
		break;

	case B_GET_ICON:
		DPRINTF_INFO ((MY_ID "ioctl(%d, B_GET_ICON)\n", number));
		ret = get_icon (my_dev, (device_icon *) arg);
		break;

	case B_FLUSH_DRIVE_CACHE:
		DPRINTF_INFO ((MY_ID "ioctl(%d, B_FLUSH_DRIVE_CACHE)\n", number));
		ret = do_synchronize_cache (my_dev, 0, 0 /* all blocks */);
		break;

	case B_FORMAT_DEVICE:
		DPRINTF_INFO ((MY_ID "ioctl(%d, B_FORMAT_DEVICE)\n", number));
		break;
		
	case B_EJECT_DEVICE:
		DPRINTF_INFO ((MY_ID "ioctl(%d, B_EJECT_DEVICE)\n", number));
		ret = do_start_stop_unit (my_dev, 2);
		break;

	case B_LOAD_MEDIA:
		DPRINTF_INFO ((MY_ID "ioctl(%d, B_LOAD_MEDIA)\n", number));
		ret = do_start_stop_unit (my_dev, 3);
		break;

	/* functions for CD-ROM drives */
	case B_SCSI_GET_TOC:
		DPRINTF_INFO ((MY_ID "ioctl(%d, B_SCSI_GET_TOC)\n", number));
		if (my_dev->inq.device_type == B_SCSI_CD ||
		    my_dev->inq.device_type == B_SCSI_RMC_CD)
		{
			ret = do_read_toc (my_dev, (void *) arg, sizeof (scsi_toc), 
				MODE_MSF, 0);
		}
		break;

	case B_SCSI_PLAY_TRACK:
		DPRINTF_INFO ((MY_ID "ioctl(%d, B_SCSI_PLAY_TRACK)\n", number));
		ret = do_play_audio_track_index (my_dev, (scsi_play_track *) arg);
		break;

	case B_SCSI_PLAY_POSITION:
		DPRINTF_INFO ((MY_ID "ioctl(%d, B_SCSI_PLAY_POSITION)\n", number));
		ret = do_play_audio_msf (my_dev, (scsi_play_position *) arg);
		break;

	case B_SCSI_PAUSE_AUDIO:
	case B_SCSI_RESUME_AUDIO:
		if (op == B_SCSI_PAUSE_AUDIO)
			DPRINTF_INFO ((MY_ID "ioctl(%d, B_SCSI_PAUSE_AUDIO)\n", number));
		else
			DPRINTF_INFO ((MY_ID "ioctl(%d, B_SCSI_RESUME_AUDIO)\n", number));
		ret = do_pause_resume (my_dev, op == B_SCSI_RESUME_AUDIO ? 1 : 0);
		break;

	case B_SCSI_GET_POSITION:
		DPRINTF_INFO ((MY_ID "ioctl(%d, B_SCSI_GET_POSITION)\n", number));
		ret = do_read_sub_channel (my_dev, (void *) arg, sizeof (scsi_position), 
			MODE_MSF, 0, 0x01);
		break;

	case B_SCSI_SET_VOLUME:
		DPRINTF_INFO ((MY_ID "ioctl(%d, B_SCSI_SET_VOLUME)\n", number));
		{
			audio_ctrl_page cur_vol;
			scsi_volume *new_vol;

			new_vol = (scsi_volume *) arg;
			ret = do_mode_sense_10 (my_dev, (void *) &cur_vol,
				sizeof (audio_ctrl_page), 0, 14);
			if (ret == B_OK)
			{
				if (new_vol->flags & 1)
					cur_vol.ports [0].channels = new_vol->port0_channel;
				if (new_vol->flags & 2)
					cur_vol.ports [0].volume   = new_vol->port0_volume;
				if (new_vol->flags & 4)
					cur_vol.ports [1].channels = new_vol->port1_channel;
				if (new_vol->flags & 8)
					cur_vol.ports [1].volume   = new_vol->port1_volume;
				if (new_vol->flags & 0x10)
					cur_vol.ports [2].channels = new_vol->port2_channel;
				if (new_vol->flags & 0x20)
					cur_vol.ports [2].volume   = new_vol->port2_volume;
				if (new_vol->flags & 0x40)
					cur_vol.ports [3].channels = new_vol->port3_channel;
				if (new_vol->flags & 0x80)
					cur_vol.ports [3].volume   = new_vol->port3_volume;
				ret = do_mode_select_10 (my_dev, (void *) &cur_vol,
					sizeof (audio_ctrl_page));
			}
		}
		break;

	case B_SCSI_GET_VOLUME:
		DPRINTF_INFO ((MY_ID "ioctl(%d, B_SCSI_GET_VOLUME)\n", number));
		{
			audio_ctrl_page cur_vol;
			scsi_volume *new_vol;

			new_vol = (scsi_volume *) arg;
			ret = do_mode_sense_10 (my_dev, (void *) &cur_vol, 
				sizeof (audio_ctrl_page), 0, 14);
			if (ret == B_OK)
			{
				new_vol->port0_channel = cur_vol.ports [0].channels;
				new_vol->port0_volume  = cur_vol.ports [0].volume;
				new_vol->port1_channel = cur_vol.ports [1].channels;
				new_vol->port1_volume  = cur_vol.ports [1].volume;
				new_vol->port2_channel = cur_vol.ports [2].channels;
				new_vol->port2_volume  = cur_vol.ports [2].volume;
				new_vol->port3_channel = cur_vol.ports [3].channels;
				new_vol->port3_volume  = cur_vol.ports [3].volume;
			}
		}
		break;

	case B_SCSI_READ_CD:
		DPRINTF_INFO ((MY_ID "ioctl(%d, B_SCSI_READ_CD)\n", number));
		{
			scsi_read_cd *read_cd;
			scsi_play_position ppos;
			uint32 end_lba;

#define	bcd2bin(x)	(((x) >> 4) * 10 + ((x) & 15))
#define	bin2bcd(x)	((((x) / 10) << 4) | (x) % 10)

			read_cd = (scsi_read_cd *) arg;
			ppos.start_m = read_cd->start_m;
			ppos.start_s = read_cd->start_s;
			ppos.start_f = read_cd->start_f;
			end_lba = 
				(bcd2bin (read_cd->start_m) * 60 + 
				 bcd2bin (read_cd->start_s)) * 75 + 
				 bcd2bin (read_cd->start_f) +
				(bcd2bin (read_cd->length_m) * 60 + 
				 bcd2bin (read_cd->length_s)) * 75 + 
				 bcd2bin (read_cd->length_f) - 1;
			ppos.end_m = bin2bcd ((end_lba / 75) / 60);
			ppos.end_s = bin2bcd ((end_lba / 75) % 60);
			ppos.end_f = bin2bcd ( end_lba % 75      );
			ret = do_read_cd_msf (my_dev, read_cd->buffer, 
				read_cd->buffer_length, &ppos);
		}
		break;

	case B_SCSI_SCAN:
		DPRINTF_INFO ((MY_ID "ioctl(%d, B_SCSI_SCAN)\n", number));
		{
			scsi_scan *scan;
			scan = (scsi_scan *) arg;
			if (scan->direction != 0)
			{
				scsi_position pos;
				ret = do_read_sub_channel (my_dev, (void *) &pos, 
					sizeof (scsi_position), MODE_MSF, 0, 0x01);
				if (ret == B_OK)
				{
					int dir = scan->direction > 0 ? DIR_FWD : DIR_REW;
					scsi_play_position ppos;
					ppos.start_m = pos.position [ 9];
					ppos.start_s = pos.position [10];
					ppos.start_f = pos.position [11];
					ret = do_scan_msf (my_dev, dir, &ppos);
				}
				break;
			}
		}
		/* else fall thru */

	case B_SCSI_STOP_AUDIO:
		DPRINTF_INFO ((MY_ID "ioctl(%d, B_SCSI_STOP_AUDIO)\n", number));
		ret = do_stop_play_scan (my_dev);
		/* device may not support STOP PLAY/SCAN (SCSI-3) */
		if (ret == B_DEV_INVALID_IOCTL)
		{
			DPRINTF_ERR ((MY_ID MY_WARN "STOP PLAY/SCAN unsupported\n"));
			ret = do_start_stop_unit (my_dev, 0);
		}
		break;

	case B_SCSI_PREVENT_ALLOW:
		DPRINTF_INFO ((MY_ID "ioctl(%d, B_SCSI_PREVENT_ALLOW)\n", number));
		ret = do_prevent_allow_medium_removal (my_dev, *(int *) arg);
		break;

	case B_SCSI_DATA_MODE:
		/* unsupported */

	case B_GET_DEVICE_SIZE:
	case B_SET_DEVICE_SIZE:
		/* obsolete */

	case B_SET_BLOCKING_IO:
	case B_SET_NONBLOCKING_IO:
	case B_GET_READ_STATUS:
	case B_GET_WRITE_STATUS:
	case B_SET_UNINTERRUPTABLE_IO:
	case B_SET_INTERRUPTABLE_IO:
		/* unsupported */

	case B_GET_BIOS_GEOMETRY:
	case B_GET_BIOS_DRIVE_ID:
		/* N/A */

	case B_GET_PARTITION_INFO:
	case B_SET_PARTITION:
	case B_GET_DRIVER_FOR_DEVICE:
		/* implemented in devfs */

	default:
		DPRINTF_ERR ((MY_ID MY_ERR "ioctl(%d, %d): unsupported\n", 
			number, (int)op));
		ret = B_DEV_INVALID_IOCTL;
		break;
	}

#ifdef DATA_TOGGLE_WORKAROUND
	do_finish (my_dev);
#endif
	return ret;
}

/*
	my_device_close
*/

static status_t my_device_close (void *cookie)
{
	my_open_info *my_fd = cookie;
	my_device_info *my_dev;

	assert (cookie != NULL && my_fd->my_dev != NULL);
	my_dev = my_fd->my_dev;
	DPRINTF_INFO ((MY_ID "close(%d)\n", my_dev->number));
	my_fd->status = B_ERROR;

	acquire_sem (my_dev->sem_lock);
	if (my_dev->open_fds == my_fd)
		my_dev->open_fds = my_fd->next;
	else
	{
		my_open_info *p;
		for (p = my_dev->open_fds; p != NULL; p = p->next)
		{
			if (p->next == my_fd)
			{
				p->next = my_fd->next;
				break;
			}
		}
	}
	--my_dev->open;
	release_sem (my_dev->sem_lock);

	return B_OK;
}

/*
	my_device_free
*/

static status_t my_device_free (void *cookie)
{
	my_open_info *my_fd = cookie;
	my_device_info *my_dev;

	assert (cookie != NULL && my_fd->my_dev != NULL);
	my_dev = my_fd->my_dev;
	DPRINTF_INFO ((MY_ID "free(%d)\n", my_dev->number));
	free (my_fd);
	if (my_dev->open > 0)
		DPRINTF_INFO ((MY_ID "%d opens left\n", my_dev->open));
	else if (!my_dev->active)
	{
		DPRINTF_INFO ((MY_ID "removed %d\n", my_dev->number));
		remove_device (my_dev);
	}
	return B_OK;
}


/*
	device hooks
*/

int32 api_version = B_CUR_DRIVER_API_VERSION;

device_hooks my_device_hooks = {
	my_device_open, 			/* -> open entry point */
	my_device_close, 			/* -> close entry point */
	my_device_free,			/* -> free cookie */
	my_device_control, 		/* -> control entry point */
	my_device_read,			/* -> read entry point */
	my_device_write,			/* -> write entry point */
	NULL, NULL,
	my_device_readv,
	my_device_writev
};

/*
	find_device
*/

device_hooks *find_device(const char *name)
{
	assert (name != NULL);
	DPRINTF_INFO ((MY_ID "find_device(%s)\n", name));
	return &my_device_hooks;
}

