/*
	USB mass storage driver for BeOS R5
	Copyright (C) 2000 ITO, Takayuki
	All rights reserved
*/

#include <drivers/KernelExport.h>
#include <drivers/Drivers.h>
#include <support/Errors.h>
#include <drivers/USB.h>
#include <drivers/bus_manager.h>
#include <storage/Mime.h>
#include <device/scsi.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <iovec.h>
#include "usb_mass.h"
#include "scsi.h"

#define	MY_ID	"\033[34musb_mass:\033[m "
#define	MY_ERR	"\033[31merror:\033[m "
#define	MY_WARN	"\033[31mwarning:\033[m "
#define	assert(x) \
	((x) ? 0 : dprintf (MY_ID "assertion failed at " __FILE__ ", line %d\n", __LINE__))

#define	BUF_SIZ	32768	/* max size of a readv/writev call */

#define	DEFAULT_CONFIGURATION	1
#define	DEFAULT_INTERFACE	1

/*
	BeOS R5 USBD seems to neglect to reset internal data toggle
	upon set_configuration or clear_feature(endpoint_halt)
*/
#define	DATA_TOGGLE_WORKAROUND

struct transport_hooks;
struct my_device_info;
struct my_open_info;

typedef struct my_device_info
{
	/* list structure */
	struct my_device_info *next;

	/* maintain device */
	sem_id sem_cb;
	sem_id sem_lock;
	area_id buffer_area;
	void *buffer;
	const usb_device *dev;
	int number;
	int subclass, protocol;
	const struct transport_hooks *hooks;
	const usb_endpoint_info *bulkInEpt, *bulkOutEpt, *intrEpt;
	size_t sector_size;
	std_inquiry_data inq;
	int cdblen;	/* 12 (fixed) or 0 (command dependent) */

	bool active;
	int open;
	struct my_open_info *open_fds;

	/* workarea for transfer */
	int usbd_status, bus_status, cmd_status;
	size_t bulk_length;
	size_t actual_length;		/* internal use */
#ifdef DATA_TOGGLE_WORKAROUND
	int bulkInEptToggle, bulkOutEptToggle, intrEptToggle;
#endif
} my_device_info;

typedef struct transport_hooks
{
	status_t (*init)(my_device_info *);
	status_t (*exec_cmd_nondata)(my_device_info *, const void *, size_t);
	status_t (*exec_cmd_datain)(my_device_info *, const void *, size_t,
		void *, size_t);
	status_t (*exec_cmd_dataout)(my_device_info *, const void *, size_t,
		const void *, size_t);
	status_t (*exec_cmd_datain_sg)(my_device_info *, const void *, size_t,
		const iovec *, size_t, size_t);
	status_t (*exec_cmd_dataout_sg)(my_device_info *, const void *, size_t,
		const iovec *, size_t, size_t);
	status_t (*finish)(my_device_info *);
} transport_hooks;

typedef struct unusual_dev
{
	uint16 vendor_id;
	uint16 product_id;
	uint16 device_version;
	uint8 subclass;
	uint8 protocol;
	uint32 flags;
} unusual_dev;

/* driver.c */

extern usb_module_info *usb;
extern const char *my_driver_name;
extern const char *my_base_name;


/* devmgmt.c */

my_device_info *
create_device (const usb_device *dev, const usb_interface_info *ii, const unusual_dev *un_dev);

void 
remove_device (my_device_info *my_dev);

/* devlist.c */

extern sem_id my_device_list_lock;
extern bool my_device_list_changed;

void add_device_info (my_device_info *my_dev);
void remove_device_info (my_device_info *my_dev);
my_device_info *search_device_info (int number);

extern char **my_device_names;

void alloc_device_names (void);
void free_device_names (void);
void rebuild_device_names (void);

/* cmds.c */

status_t do_reset
	(my_device_info *my_dev);
status_t do_test_unit_ready
	(my_device_info *my_dev);
status_t do_rezero_unit
	(my_device_info *my_dev);
status_t do_format_unit
	(my_device_info *my_dev);
status_t do_start_stop_unit
	(my_device_info *my_dev, int flags);
status_t do_prevent_allow_medium_removal
	(my_device_info *my_dev, int flags);
status_t do_seek_10
	(my_device_info *my_dev, uint32 lba);
status_t do_verify_10
	(my_device_info *my_dev, uint32 lba, uint16 blk);
status_t do_synchronize_cache
	(my_device_info *my_dev, uint32 lba, uint16 blk);
status_t do_play_audio_track_index
	(my_device_info *my_dev, scsi_play_track *track);
status_t do_play_audio_msf
	(my_device_info *my_dev, scsi_play_position *pos);
status_t do_pause_resume
	(my_device_info *my_dev, int flag);
status_t do_stop_play_scan
	(my_device_info *my_dev);
status_t do_scan_msf
	(my_device_info *my_dev, int dir, scsi_play_position *pos);
status_t do_scan_track
	(my_device_info *my_dev, int dir, int track);

status_t do_inquiry
	(my_device_info *my_dev, uint8 *buf, size_t len);
status_t do_request_sense
	(my_device_info *my_dev, uint8 *buf, size_t len);
status_t do_read_capacity
	(my_device_info *my_dev, uint8 *buf, size_t len);
status_t do_read_format_capacities
	(my_device_info *my_dev, uint8 *buf, size_t len);
status_t do_mode_sense_10
	(my_device_info *my_dev, uint8 *buf, size_t len, int pc, int page);
status_t do_read_10
	(my_device_info *my_dev, const iovec *vec, size_t vecsiz, uint32 lba, uint16 blk);
status_t do_read_12
	(my_device_info *my_dev, const iovec *vec, size_t vecsiz, uint32 lba, uint32 blk);
status_t do_read_toc
	(my_device_info *my_dev, uint8 *buf, size_t len, int mode, int track);
status_t do_read_sub_channel
	(my_device_info *my_dev, uint8 *buf, size_t len, int mode, int track, int format);
status_t do_read_cd_msf
	(my_device_info *my_dev, uint8 *buf, size_t len, scsi_play_position *pos);

status_t do_mode_select_10
	(my_device_info *my_dev, const uint8 *buf, size_t len);
status_t do_format_unit_ufi
	(my_device_info *my_dev, const uint8 *buf, size_t len, int track);
status_t do_write_10
	(my_device_info *my_dev, const iovec *vec, size_t vecsiz, uint32 lba, uint16 blk);
status_t do_write_12
	(my_device_info *my_dev, const iovec *vec, size_t vecsiz, uint32 lba, uint32 blk);
status_t do_write_and_verify_10
	(my_device_info *my_dev, const iovec *vec, size_t vecsiz, uint32 lba, uint16 blk);

status_t do_finish (my_device_info *my_dev);

/* cbi.c */

extern transport_hooks cbi_hooks;

/* bulkonly.c */

extern transport_hooks bulkonly_hooks;

/* usb_raw.c */

status_t
do_control (my_device_info *my_dev, int request_type, int request,
	int value, int index, size_t length, void *buf);

status_t
do_bulk_sg (my_device_info *my_dev, const usb_endpoint_info *ept, 
	const iovec *vec, size_t vecsiz, size_t length);

status_t
do_bulk (my_device_info *my_dev, const usb_endpoint_info *ept, 
	void *buf, size_t length);

status_t
do_interrupt (my_device_info *my_dev, const usb_endpoint_info *ept, 
	void *buf, size_t length);

/* icon.c */

extern const uchar 
	largeDiskIcon [], 
	miniDiskIcon [], 
	largeRemovableIcon [], 
	miniRemovableIcon [], 
	largeCDIcon [], 
	miniCDIcon [], 
	largeFloppyIcon [],
	miniFloppyIcon [];
